
import math, time
import shelve
import os
from raepy.utils.singleton import Singleton
from ..servo.servo import Servo
from raepy.servo.servo import mutex

# Distance from finger joint to midline = 0.035
class RadialGripper(object):
    __metaclass__ = Singleton
    def __init__(self):
        self._servo = Servo()
        self._shelfdir = os.path.abspath(__file__ + "/../../") + "/shelf"
        self._mutex = mutex
        self._init_shelf()


    def _init_shelf(self):
        self._mutex.acquire()
        shelf = shelve.open(self._shelfdir)
        if "fingerlength" in shelf:
            self._fingerlength = shelf["fingerlength"]
        else:
            shelf["fingerlength"] = 0.1
        
        if "radialgripper_zero_offset" in shelf:
            self._zero_offset = shelf["radialgripper_zero_offset"]
        else:
            shelf["radialgripper_zero_offset"] = 50
        
        if "radialgripper_mid_axis_offset" in shelf:
            self._mid_axis_offset = shelf["radialgripper_mid_axis_offset"]
        else:
            shelf["radialgripper_mid_axis_offset"] = 20
        shelf.close()
        self._mutex.release()


    def set_fingerlength(self,length):
        self._fingerlength = length
        shelf = shelve.open(self._shelfdir)
        shelf["fingerlength"] = length
        shelf.close()
        print("raepy : fingerlength set to {} m".format(length))

    def to(self, position, cb=None, current=1000):
        """
        30 = i * 10 
        i gear ratio
        10 : because motor accepts 1/10 °. 3600 -> 360°
        """
        self._servo.limp()
        goal_angle = 30 * (math.degrees(math.asin(((position/2)-0.035)/self._fingerlength)) + self._mid_axis_offset)
        self._servo.move_absolute_angle(int(goal_angle)+self._zero_offset, current=current, cb=cb)
     

    def grasp(self, speed=360, current=900, feedback_cb = None):
        self._servo.jog(-speed, current=current)
        time.sleep(0.02)
        while self._servo.get_motor_mode() != 'Holding':
            if feedback_cb:
                feedback_cb(self.fingertip_distance(), self._servo.actual_current())
            time.sleep(0.01)

    def openfingers(self):
        self.to(0.2)

    def release(self, feedback_cb = None):
        self._servo.move_relative_angle(800)
        self._servo.limp()

    def calibrate(self, current=700,feedback_cb = None):
        if self._fingerlength == None:
            print("raepy: set fingerlength before calling calibrating the gripper")
            return
        self._servo.jog(-50)
        while self._servo.actual_current() < current:
            time.sleep(0.01)
        self._servo.limp()
        time.sleep(1)
        #Servo.set_zero_here()
        self._mutex.acquire()
        shelf = shelve.open(self._shelfdir)
        self._zero_offset = self._servo.actual_angle()
        shelf["radialgripper_zero_offset"] = self._zero_offset

        self._mid_axis_offset = math.degrees(math.asin(0.035/self._fingerlength))
        shelf["radialgripper_mid_axis_offset"] = self._mid_axis_offset
        shelf.close()
        self._mutex.release()
        print("raepy : offset angle set to {} deg".format(self._mid_axis_offset))
        self._servo.limp()

    def fingertip_distance(self):
        """
        returns fingertip distance in meter
        """
        return 2*math.sin(math.radians((self._servo.actual_angle()-self._zero_offset)/30)*self._fingerlength)





import time
from ..utils.singleton import Singleton
from ..servo.servo import Servo


class Rack(object):
    __metaclass__ = Singleton
    def __init__(self):
        self._servo = Servo()

    def to(self,position,current=800,cb=None):
        """
        position in meter. Constant 82644 is calculated by (43,56[mm/360°] => 0.121 [mm/°] => 8.2644 [°/mm] => 8264,46 [°/m]  * 10 => 82644 * (1/10°)/m
        """
        self._servo.move_absolute_angle(int(position*82644),current=current,cb=cb)
        return True
    
    def position(self):
        return self._servo.actual_angle()/82644

    def out(self):
        self.to(0.08)

    def home(self, cb=None):
        self._servo.jog(-50)
        cnt = 0
        while self._servo.actual_current() < 700:
            time.sleep(0.02)
            cnt+=1
            if cnt > 1000:
                return "No Rack inserted"
            if cb:
                cb(self._servo.actual_angle(), self._servo.actual_current())
        self._servo.limp()
        time.sleep(1)
        self._servo.set_zero_here()
        self.to(0.005)
        self._servo.limp()
    




class SerialConnectionError(BaseException):
    """
    Exception rasied for errors in the Serial connection between ECU and Motor
    """
    def __init__(self, message="Either there is no serial connection between Motor and Controller Unit or there are driver problems"):
        self.message = message
        super().__init__(self.message)


#! /usr/bin/env python3
import time
import threading
import RPi.GPIO as GPIO

class Button(object):
    def __init__(self,pin=13):
        self.__pin = pin

        GPIO.setmode(GPIO.BOARD)
        GPIO.setup(pin, GPIO.IN, pull_up_down=GPIO.PUD_DOWN)

        self.__state = 0
        self.__prev_state = 0
        threading.Thread(target=self.__read_states).start()
        self.__rising_edge_callback = None
        self.__falling_edge_callback = None
    def __del__(self):
        GPIO.cleanup()
        

    def __read_states(self):
        """
        Read the button states continously inside an threaded loop
        """
        while True:
            self.__state = GPIO.input(self.__pin)
            if self.__state > self.__prev_state:
                self.__rising_edge()
            elif self.__state < self.__prev_state:
                self.__falling_edge()
            time.sleep(0.05)
            self.__prev_state = self.__state

    def __rising_edge(self):
        """
        Calls the callback if available
        """
        if self.__rising_edge_callback != None:
            self.__rising_edge_callback()

    def __falling_edge(self):
        """
        Calls the callback if available
        """
        if self.__falling_edge_callback != None:
            self.__falling_edge_callback()

    def register_rising_edge_cb(self,cb):
        """
        Register callback that will be triggered when the button gets clicked
        """
        self.__rising_edge_callback = cb

    def register_falling_edge_cb(self,cb):
        """
        Register callback that will be triggered when the button gets released
        """
        self.__falling_edge_callback = cb


    def state(self):
        """
        returns the current state of the button
        0 : Button pressed
        1 : Button released
        """
        return int(self.__state)
        
def pressed_event():
    print("Button pressed event")

def released_event():
    print("Button released event")

if __name__ == "__main__":

    lb = Button(pin=13)
    lb.register_rising_edge_cb(pressed_event)
    lb.register_falling_edge_cb(released_event)

    while True:
        if lb.state() == 1:
            print("Button pressed")
        else:
            print("Button released")
        time.sleep(1)



import time

from .LSS_Library_Python import lss
from .LSS_Library_Python import lssc

from serial.serialutil import SerialException

from .exceptions import SerialConnectionError

import sys
import glob
import serial
import os

import shelve

from raepy.utils.singleton import Singleton

from threading import Lock

mutex = Lock()


class Servo(object):
    __metaclass__ = Singleton
    def __init__(self, CST_LSS_Baud = lssc.LSS_DefaultBaud):

        self._CST_LSS_Baud = CST_LSS_Baud
        serial_ports_list = self.serial_ports()

        for port in serial_ports_list:
            try:
                self.init_interface(port)
            except SerialException:
                print("SerialException on port {}".format(port))
            if self.get_target_angle() != None:
                break


        self._shelfdir = os.path.abspath(__file__ + "/../../") +"/shelf"
        self._mutex = mutex
        self._mutex.acquire()
        shelf = shelve.open(self._shelfdir)
        if "Servo_offset" in shelf:
            self._offset_angle = shelf["Servo_offset"]
        else:
            shelf["Servo_offset"] = 0
            self._offset_angle = 0
        shelf.close()
        self._mutex.release()

        self._trial_counter = 0
        self._led_switcher = {
            "black": lssc.LSS_LED_Black,
            "red": lssc.LSS_LED_Red,
            "green": lssc.LSS_LED_Green,
            "blue": lssc.LSS_LED_Blue,
            "yellow": lssc.LSS_LED_Yellow,
            "cyan": lssc.LSS_LED_Cyan,
            "magenta": lssc.LSS_LED_Magenta,
            "white": lssc.LSS_LED_White
        }

        self._motormodes={
            0:'Unknown',
            1:'Limp',
            2:'Free Moving',
            3:'Accelerating',
            4:'Traveling',
            5:'Decelerating',
            6:'Holding',
            7:'Outside limits',
            8:'Stuck',
            9:'Blocked',
            10:'Safe Mode'
        }
    
    def __del__(self):
        self.limp()

    def init_interface(self, dev_string):
        lss.initBus(dev_string,self._CST_LSS_Baud)
        self._lss = lss.LSS(0)

    def serial_ports(self):
        """ Lists serial port names

            :raises EnvironmentError:
                On unsupported or unknown platforms
            :returns:
                A list of the serial ports available on the system
        """
        if sys.platform.startswith('win'):
            ports = ['COM%s' % (i + 1) for i in range(256)]
        elif sys.platform.startswith('linux') or sys.platform.startswith('cygwin'):
            # this excludes your current terminal "/dev/tty"
            ports = glob.glob('/dev/tty[A-Za-z]*')
        elif sys.platform.startswith('darwin'):
            ports = glob.glob('/dev/tty.*')
        else:
            raise EnvironmentError('Unsupported platform')

        result = []
        for port in ports:
            try:
                s = serial.Serial(port)
                s.close()
                result.append(port)
            except (OSError, serial.SerialException):
                pass
        return result


    def jog(self,speed, current=700):
        """
        speed paramter is in 1/10 degrees/sec: 3600 is maximum which is 360 deg/sec
        """
        cnt = 0
        self._lss.wheel(speed,current)
        while speed != self.get_target_speed():
            self._lss.wheel(speed)
            time.sleep(0.02)
            cnt+=1
            if cnt > 5:
                raise SerialConnectionError

    def move_absolute_angle(self, angle, current=700,cb=None):
        """
        Absolute Positioning in 1/10 degree: 500 => 50°
        angle: target angle in 1/10 °
        current: threshold where motor halt and holds
        """
    
        if abs(self.actual_angle() - angle - self._offset_angle) < 5:
            return

        initial_angle = self.actual_angle()

        self._lss.move(int(angle) - self._offset_angle, current)

        cnt = 0
        while self.get_target_angle() - self._offset_angle != angle:
            self._lss.move(int(angle) + self._offset_angle, current)
            time.sleep(0.02)
            cnt+=1
            if cnt > 5:
                raise SerialConnectionError

        while abs(self.get_target_angle() - self._offset_angle - self.actual_angle()) > 10:
            self._lss.move(int(angle) + self._offset_angle, current)
            if cb:
                cb(initial_angle, angle - self._offset_angle, self.actual_angle())
            time.sleep(0.02)

        

    def move_relative_angle(self, angle, current=1000):
        """
        Relative positioning in 1/10 degree: 500 => 50°
        """
        first_angle = self.actual_angle()
        goal_angle = first_angle + angle
        
        self._lss.moveRelative(int(angle), current)
        diff = abs(self.actual_angle() - goal_angle)

        cnt = 0
        while diff > 30:
            last_diff = diff
            diff = abs(self.actual_angle() - goal_angle)
            time.sleep(0.02)
            if last_diff - diff == 0:
                cnt+=1
            if cnt > 10:
                print("repeat")
                self._lss.moveRelative(int(angle), current)

        

    def actual_current(self):
        """
        returns the motor current in milliAmps
        """
        current = self._lss.getCurrent()
        if current is None:
            return current
        else:
            return int(current)

    def actual_angle(self):
        """
        returns the current angle as 1/10 degree. 360° > 3600
        """
        angle = self._lss.getPosition()
        cnt = 0
        while angle is None:
            angle = self._lss.getPosition()
            cnt += 1
            if cnt > 5:
                raise SerialConnectionError
        return int(angle) - self._offset_angle

    def actual_speed(self):
        """
        returns the current speed in 1/10 °/sec
        """
        speed = self._lss.getSpeed()
        if speed == None:
            return speed
        else:
            return int(speed)

    def actual_LED_color(self):
        """
        returns the current LED Color as String
        """
        color = None
        cnt = 0
        while color == None:
            color = self._lss.getColorLED()
            cnt+=1
            time.sleep(0.02)
            if cnt > 5:
                raise SerialConnectionError
        color = int(color)
        return [color_str for color_str,id in self._led_switcher.items() if id == color][0]

    def hold(self):
        """
        holds the motor with an specific torque in position
        """

        cnt = 0
        self._lss.hold()
        while self.get_motor_mode() != 'Holding':
            self._lss.hold()
            time.sleep(0.01)
            cnt+=1
            if cnt > 5:
                raise SerialConnectionError
        return True

    
    def limp(self):
        """
        Turns holding torque off
        """
        cnt = 0
        self._lss.limp()
        while self.get_motor_mode() != 'Holding':
            self._lss.hold()
            time.sleep(0.01)
            cnt += 1
            if cnt > 5:
                raise SerialConnectionError

    def set_zero_here(self,offset=0):
        """
        Set the zero Position
        """

        self._offset_angle = self.actual_angle() + self._offset_angle
        self._mutex.acquire()
        shelf = shelve.open(self._shelfdir)
        shelf["Servo_offset"] = self._offset_angle
        shelf.close()
        self._mutex.release()

        return True

    def reset(self):
        """
        Reset Motor
        """
        self._lss.reset()

    def get_motor_mode(self):
        """
        Returns the current motor mode
        """
        mode = self._lss.getStatus()
        if mode == None:
            return 'Unknown'

        return self._motormodes.get(int(mode))

    def get_target_angle(self):
        """
        returns target Angle
        """
        cnt = 0
        while True:
            pos = self._lss.getTargetPosition()
            if pos != None:
                return int(pos)
            time.sleep(0.02)
            cnt+=1
            if cnt > 10:
                raise SerialConnectionError


    def get_target_speed(self):
        """
        return speed
        """
        return int(self._lss.getTargetSpeed())

    def led_to(self, color):
        colorcode = self._led_switcher.get(color, "invalid color")
        self._lss.setColorLED(colorcode)
        while colorcode != int(self._lss.getColorLED()):
            self.led_to(color)
            time.sleep(0.01)

    def _continue_if_motion_detected(self):
        cnt = 0
        while self.get_motor_mode() != 'Accelerating':
            time.sleep(0.01)
            cnt += 1
            if cnt > 5:
                raise SerialConnectionError



import threading
import subprocess
import atexit
import os
import signal
import time

from ..utils.singleton import Singleton

class VacuumModule(object):
    __metaclass__ = Singleton
    def __init__(self):
        print("Vacuum Module init")
        try:
            os.mkfifo("/tmp/vacmodstate")
            os.mkfifo("/tmp/vacmodcmd")
        except:
            pass
        
        file = os.path.abspath(__file__ + "/../") + "/vacmod"
        self.process = subprocess.Popen("nohup {} >/dev/null 2>&1 &".format(file),shell=True)

        atexit.register(self.clean)
        self.actual_state = None
        self.__sucked_cb = None
        self.__lost_cb = None
        time.sleep(0.1)
        state_reader = threading.Thread(target=self.__read_state)
        state_reader.start()
        
    

    def clean(self):
        print("Cleanup!")
        os.killpg(os.getpgid(self.process.pid), signal.SIGTERM)

    def suck(self, sucked_cb=None, lost_cb=None):
        with open("/tmp/vacmodcmd", "w") as startcmd:
            startcmd.write("START")

        if sucked_cb:
            self.__sucked_cb = sucked_cb
        if lost_cb:
            self.__lost_cb = lost_cb

    def warmup(self):
        with open("/tmp/vacmodcmd", "w") as startcmd:
            startcmd.write("WARMUP")
        time.sleep(1)
        while self.current_state() == "WARMUP":
            time.sleep(0.5)

        
    def release(self):
        with open("/tmp/vacmodcmd", "w") as stopcmd:
            stopcmd.write("STOP\n")

    def sucked(self):
        return self.current_state() == "SUCKED"

    def current_state(self):
        return self.actual_state

    def __read_state(self):
        while True:
            with open('/tmp/vacmodstate') as fifo:
                for state in fifo:
                    self.actual_state = state.rstrip('\x00')
                    if self.actual_state == "SUCKED" and self.__sucked_cb != None:
                        self.__sucked_cb(state)
                    elif self.actual_state == "LOST" and self.__lost_cb != None:
                        self.__lost_cb(state)
                    


    def print_state(self,state):
        print("{}".format(state),end="\n")


                    




###############################################################################
#	Author:			Sebastien Parent-Charette (support@robotshop.com)
#	Version:		1.0.0
#	Licence:		LGPL-3.0 (GNU Lesser General Public License version 3)
#	
#	Desscription:	An example using the LSS and the Python module.
###############################################################################

# Import required liraries
import time
import serial

# Import LSS library
import lss
import lss_const as lssc

# Constants
#CST_LSS_Port = "/dev/ttyUSB0"		# For Linux/Unix platforms
CST_LSS_Port = "COM230"				# For windows platforms
CST_LSS_Baud = lssc.LSS_DefaultBaud

# Create and open a serial port
lss.initBus(CST_LSS_Port, CST_LSS_Baud)

# Create LSS objects
myLSS1 = lss.LSS(0)
myLSS2 = lss.LSS(1)

#myLSS1.move(-300)
#myLSS2.move(300)

print("myLSS1; position = " + str(myLSS1.getPosition()))
print("myLSS2; position = " + str(myLSS2.getPosition()))
print("myLSS1; model = " + str(myLSS1.getModel()))

#myLSS1.reset()
#myLSS2.reset()

# Destroy objects
del myLSS1
del myLSS2

# Destroy the bus
lss.closeBus()

### EOF #######################################################################



from .servo.servo import Servo 
from .radialgripper.radialgripper import RadialGripper 
from .rack.rack import Rack 
from .vacmod.vacuummodule import VacuumModule
from .ledbutton.ledbutton import LedButton



###############################################################################
#	Author:			Sebastien Parent-Charette (support@robotshop.com)
#	Version:		1.0.0
#	Licence:		LGPL-3.0 (GNU Lesser General Public License version 3)
#	
#	Desscription:	A library that makes using the LSS simple.
#					Offers support for most Python platforms.
#					Uses the Python serial library (pySerial).
###############################################################################

### List of constants

# Bus communication
LSS_DefaultBaud = 115200
LSS_MaxTotalCommandLength = (30 + 1)	# ex: #999XXXX-2147483648\r Adding 1 for end string char (\0)
#										# ex: #999XX000000000000000000\r
LSS_Timeout = 100						# in ms
LSS_CommandStart = "#"
LSS_CommandReplyStart = "*"
LSS_CommandEnd = "\r"
LSS_FirstPositionDisabled = "DIS"

# LSS constants
LSS_ID_Default = 0
LSS_ID_Min = 0
LSS_ID_Max = 250
LSS_Mode255ID = 255
LSS_BroadcastID = 254

# Read/write status
LSS_CommStatus_Idle = 0
LSS_CommStatus_ReadSuccess = 1
LSS_CommStatus_ReadTimeout = 2
LSS_CommStatus_ReadWrongID = 3
LSS_CommStatus_ReadWrongIdentifier = 4
LSS_CommStatus_ReadWrongFormat = 5
LSS_CommStatus_ReadNoBus = 6
LSS_CommStatus_ReadUnknown = 7
LSS_CommStatus_WriteSuccess = 8
LSS_CommStatus_WriteNoBus = 9
LSS_CommStatus_WriteUnknown = 10

# LSS status
LSS_StatusUnknown = 0
LSS_StatusLimp = 1
LSS_StatusFreeMoving = 2
LSS_StatusAccelerating = 3
LSS_StatusTravelling = 4
LSS_StatusDecelerating = 5
LSS_StatusHolding = 6
LSS_StatusOutsideLimits = 7
LSS_StatusStuck = 8				#cannot move at current speed setting
LSS_StatusBlocked = 9			#same as stuck but reached maximum duty and still can't move
LSS_StatusSafeMode = 10
LSS_StatusLast = 11

# LSS models
LSS_ModelHighTorque = 0
LSS_ModelStandard = 1
LSS_ModelHighSpeed = 2
LSS_ModelUnknown = 3

LSS_ModelHT1 = "LSS-HT1"
LSS_ModelST1 = "LSS-ST1"
LSS_ModelHS1 = "LSS-HS1"

# Parameter for query
LSS_QuerySession = 0
LSS_QueryConfig = 1
LSS_QueryInstantaneousSpeed = 2
LSS_QueryTargetTravelSpeed = 3

# Parameter for setter
LSS_SetSession = 0
LSS_SetConfig = 1

# Parameter for Serial/RC mode change
LSS_ModeSerial = 0
LSS_ModePositionRC = 1
LSS_ModeWheelRC = 2

# Parameter for gyre direction
LSS_GyreClockwise = 1
LSS_GyreCounterClockwise = -1

# LED colors
LSS_LED_Black = 0
LSS_LED_Red = 1
LSS_LED_Green = 2
LSS_LED_Blue = 3
LSS_LED_Yellow = 4
LSS_LED_Cyan = 5
LSS_LED_Magenta = 6
LSS_LED_White = 7

# Commands - actions
LSS_ActionReset = "RESET"
LSS_ActionLimp = "L"
LSS_ActionHold = "H"
LSS_ActionParameterTime = "T"
LSS_ActionParameterSpeed = "S"
LSS_ActionMove = "D"
LSS_ActionMoveRelative = "MD"
LSS_ActionWheel = "WD"
LSS_ActionWheelRPM = "WR"

LSS_CurrentHaltHold = "CH"
LSS_CurrentLimpHold = "CL"

# Commands - actions settings
LSS_ActionOriginOffset = "O"
LSS_ActionAngularRange = "AR"
LSS_ActionMaxSpeed = "SD"
LSS_ActionMaxSpeedRPM = "SR"
LSS_ActionColorLED = "LED"
LSS_ActionGyreDirection = "G"

# Commands - actions advanced settings
LSS_ActionAngularStiffness = "AS"
LSS_ActionAngularHoldingStiffness = "AH"
LSS_ActionAngularAcceleration = "AA"
LSS_ActionAngularDeceleration = "AD"
LSS_ActionEnableMotionControl = "EM"

# Commands - queries
LSS_QueryStatus = "Q"
LSS_QueryOriginOffset = "QO"
LSS_QueryAngularRange = "QAR"
LSS_QueryPositionPulse = "QP"
LSS_QueryPosition = "QD"
LSS_QueryTargetPosition = "QDT"
LSS_QuerySpeed = "QWD"
LSS_QueryTargetSpeed = "QVT"
LSS_QuerySpeedRPM = "QWR"
LSS_QuerySpeedPulse = "QS"
LSS_QueryMaxSpeed = "QSD"
LSS_QueryMaxSpeedRPM = "QSR"
LSS_QueryColorLED = "QLED"
LSS_QueryGyre = "QG"
LSS_QueryID = "QID"
LSS_QueryBaud = "QB"
LSS_QueryFirstPosition = "QFD"
LSS_QueryModelString = "QMS"
LSS_QuerySerialNumber = "QN"
LSS_QueryFirmwareVersion = "QF"
LSS_QueryVoltage = "QV"
LSS_QueryTemperature = "QT"
LSS_QueryCurrent = "QC"

# Commands - queries advanced
LSS_QueryAngularStiffness = "QAS"
LSS_QueryAngularHoldingStiffness = "QAH"
LSS_QueryAngularAcceleration = "QAA"
LSS_QueryAngularDeceleration = "QAD"
LSS_QueryEnableMotionControl = "QEM"
LSS_QueryBlinkingLED = "QLB"

# Commands - configurations
LSS_ConfigID = "CID"
LSS_ConfigBaud = "CB"
LSS_ConfigOriginOffset = "CO"
LSS_ConfigAngularRange = "CAR"
LSS_ConfigMaxSpeed = "CSD"
LSS_ConfigMaxSpeedRPM = "CSR"
LSS_ConfigColorLED = "CLED"
LSS_ConfigGyreDirection = "CG"
LSS_ConfigFirstPosition = "CFD"
LSS_ConfigMode = "CRC"

# Commands - configurations advanced
LSS_ConfigAngularStiffness = "CAS"
LSS_ConfigAngularHoldingStiffness = "CAH"
LSS_ConfigAngularAcceleration = "CAA"
LSS_ConfigAngularDeceleration = "CAD"
LSS_ConfigBlinkingLED = "CLB"

### EOF #######################################################################



###############################################################################
#	Author:			Sebastien Parent-Charette (support@robotshop.com)
#	Version:		1.0.0
#	Licence:		LGPL-3.0 (GNU Lesser General Public License version 3)
#	
#	Desscription:	A library that makes using the LSS simple.
#					Offers support for most Python platforms.
#					Uses the Python serial library (pySerial).
###############################################################################

### Import required liraries
import re
import serial
from math import sqrt, atan, acos, fabs

### Import constants
from raepy.servo.LSS_Library_Python import lss_const as lssc

### Class functions
def initBus(portName, portBaud):
	LSS.bus = serial.Serial(portName, portBaud)
	LSS.bus.timeout = 0.1

def closeBus():
	if LSS.bus is not None:
		LSS.bus.close()
		del LSS.bus

# Write with a an optional parameter
def genericWrite(id, cmd, param = None):
	if LSS.bus is None:
		return False
	if param is None:
		LSS.bus.write((lssc.LSS_CommandStart + str(id) + cmd + lssc.LSS_CommandEnd).encode())
	else:
		LSS.bus.write((lssc.LSS_CommandStart + str(id) + cmd + str(param) + lssc.LSS_CommandEnd).encode())
	return True

# Read an integer result
def genericRead_Blocking_int(id, cmd):
	if LSS.bus is None:
		return None
	try:
		# Get start of packet and discard header and everything before
		c = LSS.bus.read()
		while (c.decode("utf-8") != lssc.LSS_CommandReplyStart):
			c = LSS.bus.read()
			if(c.decode("utf-8") == ""):
				break
		# Get packet
		data = LSS.bus.read_until(lssc.LSS_CommandEnd.encode('utf-8')) #Otherwise (without ".encode('utf-8')") the received LSS_CommandEnd is not recognized by read_until, making it wait until timeout.
		# Parse packet
		matches = re.match("(\d{1,3})([A-Z]{1,4})(-?\d{1,18})", data.decode("utf-8"), re.I)
		# Check if matches are found
		if(matches is None):
			return(None)
		if((matches.group(1) is None) or (matches.group(2) is None) or (matches.group(3) is None)):
			return(None)
		# Get values from match
		readID = matches.group(1)
		readIdent = matches.group(2)
		readValue = matches.group(3)
		# Check id
		if(readID != str(id)):
			return(None)
		# Check identifier
		if(readIdent != cmd):
			return(None)
	except:
		return(None)
	# return value
	return(readValue)

# Read a string result
#@classmethod
def genericRead_Blocking_str(id, cmd, numChars):
	if LSS.bus is None:
		return None
	if LSS.bus is None:
		return None
	try:
		# Get start of packet and discard header and everything before
		c = LSS.bus.read()
		while (c.decode("utf-8") != lssc.LSS_CommandReplyStart):
			c = LSS.bus.read()
			if(c.decode("utf-8") == ""):
				break
		# Get packet
		data = LSS.bus.read_until(lssc.LSS_CommandEnd.encode('utf-8')) #Otherwise (without ".encode('utf-8')") the received LSS_CommandEnd is not recognized by read_until, making it wait until timeout.
		data = (data[:-1])
		# Parse packet
		matches = re.match("(\d{1,3})([A-Z]{1,4})(.{" + str(numChars) + "})", data.decode("utf-8"), re.I)
		# Check if matches are found
		if(matches is None):
			return(None)
		if((matches.group(1) is None) or (matches.group(2) is None) or (matches.group(3) is None)):
			return(None)
		# Get values from match
		readID = matches.group(1)
		readIdent = matches.group(2)
		readValue = matches.group(3)
		# Check id
		if(readID != str(id)):
			return(None)
		# Check identifier
		if(readIdent != cmd):
			return(None)
	except:
		return(None)
	# return value
	return(readValue)

class LSS:
	# Class attribute
	bus = None
	
	### Constructor
	def __init__(self, id = 0):
		self.servoID = id
	
	### Attributes
	servoID = 0
	
	### Functions
	#> Actions
	def reset(self):
		return (genericWrite(self.servoID, lssc.LSS_ActionReset))
	
	def limp(self):
		return (genericWrite(self.servoID, lssc.LSS_ActionLimp))
	
	def hold(self):
		return (genericWrite(self.servoID, lssc.LSS_ActionHold))
	
	def move(self, pos, current=None):
		if current == None:
			return (genericWrite(self.servoID, lssc.LSS_ActionMove, pos))
		else:
			param = str(pos) + lssc.LSS_CurrentHaltHold + str(current)
			return (genericWrite(self.servoID, lssc.LSS_ActionMove, param))

	
	def moveRelative(self, delta, current=None):
		if current == None: 
			return (genericWrite(self.servoID, lssc.LSS_ActionMoveRelative, delta))
		else:
			param = str(delta) + lssc.LSS_CurrentHaltHold + str(current)
			return (genericWrite(self.servoID, lssc.LSS_ActionMoveRelative, param))
	
	def wheel(self, speed, current=None):
		if current == None:
			return (genericWrite(self.servoID, lssc.LSS_ActionWheel, speed))
		else:
			param = str(speed) + lssc.LSS_CurrentHaltHold + str(current)
			return (genericWrite(self.servoID, lssc.LSS_ActionWheel, param))

	def wheelRPM(self, rpm):
		return (genericWrite(self.servoID, lssc.LSS_ActionWheelRPM, rpm))
	
	#> Queries
	#def getID(self):
	#def getBaud(self):
	
	def getStatus(self):
		genericWrite(self.servoID, lssc.LSS_QueryStatus)
		return (genericRead_Blocking_int(self.servoID, lssc.LSS_QueryStatus))
	
	def getOriginOffset(self, queryType = lssc.LSS_QuerySession):
		genericWrite(self.servoID, lssc.LSS_QueryOriginOffset, queryType)
		return (genericRead_Blocking_int(self.servoID, lssc.LSS_QueryOriginOffset))
	
	def getAngularRange(self, queryType = lssc.LSS_QuerySession):
		genericWrite(self.servoID, lssc.LSS_QueryAngularRange, queryType)
		return (genericRead_Blocking_int(self.servoID, lssc.LSS_QueryAngularRange))
	
	def getPositionPulse(self):
		genericWrite(self.servoID, lssc.LSS_QueryPositionPulse)
		return (genericRead_Blocking_int(self.servoID, lssc.LSS_QueryPositionPulse))
	
	def getPosition(self):
		genericWrite(self.servoID, lssc.LSS_QueryPosition)
		return (genericRead_Blocking_int(self.servoID, lssc.LSS_QueryPosition))
	
	def getSpeed(self):
		genericWrite(self.servoID, lssc.LSS_QuerySpeed)
		return (genericRead_Blocking_int(self.servoID, lssc.LSS_QuerySpeed))
	
	def getSpeedRPM(self):
		genericWrite(self.servoID, lssc.LSS_QuerySpeedRPM)
		return (genericRead_Blocking_int(self.servoID, lssc.LSS_QuerySpeedRPM))
	
	def getSpeedPulse(self):
		genericWrite(self.servoID, lssc.LSS_QuerySpeedPulse)
		return (genericRead_Blocking_int(self.servoID, lssc.LSS_QuerySpeedPulse))
	
	def getMaxSpeed(self, queryType = lssc.LSS_QuerySession):
		genericWrite(self.servoID, lssc.LSS_QueryMaxSpeed, queryType)
		return (genericRead_Blocking_int(self.servoID, lssc.LSS_QueryMaxSpeed))
	
	def getMaxSpeedRPM(self, queryType = lssc.LSS_QuerySession):
		genericWrite(self.servoID, lssc.LSS_QueryMaxSpeedRPM, queryType)
		return (genericRead_Blocking_int(self.servoID, lssc.LSS_QueryMaxSpeedRPM))
	
	def getColorLED(self, queryType = lssc.LSS_QuerySession):
		genericWrite(self.servoID, lssc.LSS_QueryColorLED, queryType)
		return (genericRead_Blocking_int(self.servoID, lssc.LSS_QueryColorLED))
	
	def getGyre(self, queryType = lssc.LSS_QuerySession):
		genericWrite(self.servoID, lssc.LSS_QueryGyre, queryType)
		return (genericRead_Blocking_int(self.servoID, lssc.LSS_QueryGyre))
	
	# returns 0 if "DIS"
	def getFirstPosition(self):
		genericWrite(self.servoID, lssc.LSS_QueryFirstPosition)
		return (genericRead_Blocking_int(self.servoID, lssc.LSS_QueryFirstPosition))
	
	# returns true/false based on if QFD returns "DIS" (= False)
	def getIsFirstPositionEnabled(self):
		genericWrite(self.servoID, lssc.LSS_QueryFirstPosition)
		return (genericRead_Blocking_int(self.servoID, lssc.LSS_QueryFirstPosition) is not None)
	
	def getModel(self):
		genericWrite(self.servoID, lssc.LSS_QueryModelString)
		return (genericRead_Blocking_str(self.servoID, lssc.LSS_QueryModelString, 7))
	
	def getSerialNumber(self):
		genericWrite(self.servoID, lssc.LSS_QuerySerialNumber)
		return (genericRead_Blocking_int(self.servoID, lssc.LSS_QuerySerialNumber))
	
	def getFirmwareVersion(self):
		genericWrite(self.servoID, lssc.LSS_QueryFirmwareVersion)
		return (genericRead_Blocking_int(self.servoID, lssc.LSS_QueryFirmwareVersion))
	
	def getVoltage(self):
		genericWrite(self.servoID, lssc.LSS_QueryVoltage)
		return (genericRead_Blocking_int(self.servoID, lssc.LSS_QueryVoltage))
	
	def getTemperature(self):
		genericWrite(self.servoID, lssc.LSS_QueryTemperature)
		return (genericRead_Blocking_int(self.servoID, lssc.LSS_QueryTemperature))
	
	def getCurrent(self):
		genericWrite(self.servoID, lssc.LSS_QueryCurrent)
		return (genericRead_Blocking_int(self.servoID, lssc.LSS_QueryCurrent))
	
	#> Queries (advanced)
	def getAngularStiffness(self, queryType = lssc.LSS_QuerySession):
		genericWrite(self.servoID, lssc.LSS_QueryAngularStiffness, queryType)
		return (genericRead_Blocking_int(self.servoID, lssc.LSS_QueryAngularStiffness))
	
	def getAngularHoldingStiffness(self, queryType = lssc.LSS_QuerySession):
		genericWrite(self.servoID, lssc.LSS_QueryAngularHoldingStiffness, queryType)
		return (genericRead_Blocking_int(self.servoID, lssc.LSS_QueryAngularHoldingStiffness))
	
	def getAngularAcceleration(self, queryType = lssc.LSS_QuerySession):
		genericWrite(self.servoID, lssc.LSS_QueryAngularAcceleration, queryType)
		return (genericRead_Blocking_int(self.servoID, lssc.LSS_QueryAngularAcceleration))
	
	def getAngularDeceleration(self, queryType = lssc.LSS_QuerySession):
		genericWrite(self.servoID, lssc.LSS_QueryAngularDeceleration, queryType)
		return (genericRead_Blocking_int(self.servoID, lssc.LSS_QueryAngularDeceleration))
	
	def getIsMotionControlEnabled(self):
		genericWrite(self.servoID, lssc.LSS_QueryEnableMotionControl)
		return (genericRead_Blocking_int(self.servoID, lssc.LSS_QueryEnableMotionControl))
	
	def getBlinkingLED(self):
		genericWrite(self.servoID, lssc.LSS_QueryBlinkingLED)
		return (genericRead_Blocking_int(self.servoID, lssc.LSS_QueryBlinkingLED))

	# extensions
	
	def getTargetPosition(self):
		genericWrite(self.servoID, lssc.LSS_QueryTargetPosition)
		return (genericRead_Blocking_int(self.servoID, lssc.LSS_QueryTargetPosition))

	def getTargetSpeed(self):
		genericWrite(self.servoID, lssc.LSS_QueryTargetSpeed)
		return (genericRead_Blocking_int(self.servoID, lssc.LSS_QueryTargetSpeed))		
	
	
	#> Configs
	def setOriginOffset(self, pos, setType = lssc.LSS_SetSession):
		if setType == lssc.LSS_SetSession:
			return (genericWrite(self.servoID, lssc.LSS_ActionOriginOffset, pos))
		elif setType == lssc.LSS_SetConfig:
			return (genericWrite(self.servoID, lssc.LSS_ConfigOriginOffset, pos))
	
	def setAngularRange(self, delta, setType = lssc.LSS_SetSession):
		if setType == lssc.LSS_SetSession:
			return (genericWrite(self.servoID, lssc.LSS_ActionAngularRange, delta))
		elif setType == lssc.LSS_SetConfig:
			return (genericWrite(self.servoID, lssc.LSS_ConfigAngularRange, delta))
	
	def setMaxSpeed(self, speed, setType = lssc.LSS_SetSession):
		if setType == lssc.LSS_SetSession:
			return (genericWrite(self.servoID, lssc.LSS_ActionMaxSpeed, speed))
		elif setType == lssc.LSS_SetConfig:
			return (genericWrite(self.servoID, lssc.LSS_ConfigMaxSpeed, speed))
	
	def setMaxSpeedRPM(self, rpm, setType = lssc.LSS_SetSession):
		if setType == lssc.LSS_SetSession:
			return (genericWrite(self.servoID, lssc.LSS_ActionMaxSpeedRPM, rpm))
		elif setType == lssc.LSS_SetConfig:
			return (genericWrite(self.servoID, lssc.LSS_ConfigMaxSpeedRPM, rpm))
	
	def setColorLED(self, color, setType = lssc.LSS_SetSession):
		if setType == lssc.LSS_SetSession:
			return (genericWrite(self.servoID, lssc.LSS_ActionColorLED, color))
		elif setType == lssc.LSS_SetConfig:
			return (genericWrite(self.servoID, lssc.LSS_ConfigColorLED, color))
	
	def setGyre(self, gyre, setType = lssc.LSS_SetSession):
		if setType == lssc.LSS_SetSession:
			return (genericWrite(self.servoID, lssc.LSS_ActionGyreDirection, gyre))
		elif setType == lssc.LSS_SetConfig:
			return (genericWrite(self.servoID, lssc.LSS_ConfigGyreDirection, gyre))
	
	def setFirstPosition(self, pos):
		return (genericWrite(self.servoID, lssc.LSS_ConfigFirstPosition, pos))
	
	def clearFirstPosition(self):
		return (genericWrite(self.servoID, lssc.LSS_ConfigFirstPosition))
	
	def setMode(self, mode):
		return (genericWrite(self.servoID, lssc.LSS_ConfigMode, mode))
	
	#> Configs (advanced)
	def setAngularStiffness(self, value, setType = lssc.LSS_SetSession):
		if setType == lssc.LSS_SetSession:
			return (genericWrite(self.servoID, lssc.LSS_ActionAngularStiffness, value))
		elif setType == lssc.LSS_SetConfig:
			return (genericWrite(self.servoID, lssc.LSS_ConfigAngularStiffness, value))
	
	def setAngularHoldingStiffness(self, value, setType = lssc.LSS_SetSession):
		if setType == lssc.LSS_SetSession:
			return (genericWrite(self.servoID, lssc.LSS_ActionAngularHoldingStiffness, value))
		elif setType == lssc.LSS_SetConfig:
			return (genericWrite(self.servoID, lssc.LSS_ConfigAngularHoldingStiffness, value))
	
	def setAngularAcceleration(self, value, setType = lssc.LSS_SetSession):
		if setType == lssc.LSS_SetSession:
			return (genericWrite(self.servoID, lssc.LSS_ActionAngularAcceleration, value))
		elif setType == lssc.LSS_SetConfig:
			return (genericWrite(self.servoID, lssc.LSS_ConfigAngularAcceleration, value))
	
	def setAngularDeceleration(self, value, setType = lssc.LSS_SetSession):
		if setType == lssc.LSS_SetSession:
			return (genericWrite(self.servoID, lssc.LSS_ActionAngularDeceleration, value))
		elif setType == lssc.LSS_SetConfig:
			return (genericWrite(self.servoID, lssc.LSS_ConfigAngularDeceleration, value))
	
	def setMotionControlEnabled(self, value):
		return (genericWrite(self.servoID, lssc.LSS_ActionEnableMotionControl, value))
	
	def setBlinkingLED(self, state):
		return (genericWrite(self.servoID, lssc.LSS_ConfigBlinkingLED, state))
	
### EOF #######################################################################



#! /usr/bin/env python3

import time
import threading
import RPi.GPIO as GPIO
import numpy as np
import os

class Led(object):
    def __init__(self,pin=15):
        self.__pin = pin
        GPIO.setmode(GPIO.BOARD)
        GPIO.setup(self.__pin, GPIO.OUT)
        self.__ledpwm = GPIO.PWM(self.__pin,700)
        self.__stop_event = threading.Event()
        self.__blink_thread = threading.Thread(target=self.__blinker, args=[self.__stop_event])
        self.__hearbeat_switch = threading.Event()
        self.__heartbeat_thread = threading.Thread(target=self.__heartbeat,args=[self.__hearbeat_switch])
        self.__heartbeat_thread.start()
        ecg_file_path = os.path.abspath(__file__+"/../") + "/ECG_EngZeeMod.npy"
        self.__ecg_pattern = np.load(ecg_file_path)
        self.set_heartbeat_speed(speed=1)
        self.__blink_thread.start()
        self.__blink_freq = 1
        self.__led_intensity = 0
        self.__ledpwm.start(100)

    def __del__(self):
        GPIO.cleanup()

    def dimm(self,intensity):
        self.__led_intensity = intensity
        self.__ledpwm.ChangeDutyCycle(100-intensity)

    def on(self,intensity=100):
        """
        Switches the led ring on
        """
        if self.__led_intensity != intensity:
            self.dimm(intensity)
            self.__led_intensity = intensity

     
    def off(self):
        """
        Switches the led ring on
        """
        self.dimm(0)


    def __heartbeat(self,switch_event):
        
        #v = np.abs(nk.ecg_simulate(duration=8, sampling_rate=100, heart_rate=80))
        #normalized_ecg = (v-v.min()) / (v.max()-v.min())
        #normalized_ecg = (normalized_ecg*100).astype(int)

        while True:
            if switch_event.wait():
                for sample in  self.__ecg_pattern:
                    self.__ledpwm.ChangeDutyCycle(100-int(sample))
                    time.sleep(self.__heartbeat_speed)
                    if not switch_event.wait():
                        print("here")
                        break
            else:
                time.sleep(1)
    
    def set_heartbeat_speed(self,speed):
        """
        Speed parameter can be set in between 1 and 100
        """
        if speed > 99:
            speed = 99
        elif speed < 1:
            speed = 1

        self.__heartbeat_speed = (200-speed)/100000

    def heartbeat_on(self,speed=1):
        self.set_heartbeat_speed(speed)
        self.__ledpwm.ChangeDutyCycle(100)
        self.__hearbeat_switch.set()
    
    def heartbeat_off(self):
        self.__hearbeat_switch.clear()
        self.__ledpwm.ChangeDutyCycle(100)

    def __blink_cycle(self,t):
        self.on()
        time.sleep(t)
        self.off()
        time.sleep(t)


    def __blinker(self,event):
        while True:
            if event.wait():
                self.__blink_cycle(1/(2*self.__blink_freq))
            else:
                time.sleep(1)

    def blink(self,freq=7,cnt=7):
        """
        Blink function for the LED ring:
        Parameters:
        freq : switching frequency
        cnt : How many counts will be switched.
        If cnt == 0 then it will be switched continously
        """
        self.__stop_event.clear()

        if freq == 0:
            self.off()
            return
        
        if cnt != 0:
            for _ in range(cnt):
                self.__blink_cycle((1/(2*freq)))
        else:
            self.__blink_freq = freq
            self.__stop_event.set()
    
    def blink_on(self,freq=7):
        self.blink(freq=freq,cnt=0)

    def stop_blink(self):
        self.__stop_event.clear()
        self.off()


if __name__ == "__main__":

    led = Led()
    """    
    led.on(10)
    time.sleep(3)
    led.on(100)
    time.sleep(3)
    led.off()
    time.sleep(1)
    """

    led.heartbeat_on(speed=50)
    time.sleep(15)
    led.heartbeat_off()



#! /usr/bin/env python3

from .led import Led
from .button import Button
import time

class LedButton(object):
    def __init__(self,ledpin=15,buttonpin=13):
        self.led = Led(ledpin)
        self.button = Button(buttonpin)


if __name__ == "__main__":

    lb = LedButton()
    while True:
        print(lb.button.state())
        if lb.button.state() == 1:
            lb.led.on()
        else:
            lb.led.off()
        time.sleep(0.1)
